/*
 * $Id: disre.c,v 1.31 2005/08/07 10:11:27 spoel Exp $
 * 
 *                This source code is part of
 * 
 *                 G   R   O   M   A   C   S
 * 
 *          GROningen MAchine for Chemical Simulations
 * 
 *                        VERSION 3.2.0
 * Written by David van der Spoel, Erik Lindahl, Berk Hess, and others.
 * Copyright (c) 1991-2000, University of Groningen, The Netherlands.
 * Copyright (c) 2001-2004, The GROMACS development team,
 * check out http://www.gromacs.org for more information.

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * If you want to redistribute modifications, please consider that
 * scientific software is very special. Version control is crucial -
 * bugs must be traceable. We will be happy to consider code for
 * inclusion in the official distribution, but derived work must not
 * be called official GROMACS. Details are found in the README & COPYING
 * files - if they are missing, get the official version at www.gromacs.org.
 * 
 * To help us fund GROMACS development, we humbly ask that you cite
 * the papers on the package - you can find them in the top README file.
 * 
 * For more info, check our website at http://www.gromacs.org
 * 
 * And Hey:
 * GROningen Mixture of Alchemy and Childrens' Stories
 */
/* This file is completely threadsafe - keep it that way! */

/* Modified by Szpari for S2 restraint calculations
 * The new code will have to be put in a seperate file but for
 * testing I won't bother with persuading make to include it.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>
#include "typedefs.h"
#include "sysstuff.h"
#include "smalloc.h"
#include "macros.h"
#include "vec.h"
#include "futil.h"
#include "xvgr.h"
#include "fatal.h"
#include "bondf.h"
#include "copyrite.h"
#include "disre.h"
#include "main.h"
#include "do_fit.h" /* ZG */
#include "mdrun.h" /* ZG */


/* GZ */

void init_disres(FILE *fplog,int nfa,const t_iatom forceatoms[],
		 const t_iparams ip[],
		 const t_inputrec *ir,const t_commrec *mcr,t_fcdata *fcd)
{
  int          fa;
  t_disresdata *dd;

  dd = &(fcd->disres);

  dd->dr_weighting = ir->eDisreWeighting;
  dd->dr_fc        = ir->dr_fc;
  dd->dr_tau       = ir->dr_tau;
  if (dd->dr_tau == 0.0) {
    dd->dr_bMixed = FALSE;
    dd->ETerm = 0.0;
  } else {
    dd->dr_bMixed = ir->bDisreMixed;
    dd->ETerm = exp(-(ir->delta_t/ir->dr_tau));
  }
  dd->ETerm1 = 1.0 - dd->ETerm;
  dd->exp_min_t_tau = 1.0;
  
  dd->nr = 0;
  for(fa=0; fa<nfa; fa+=3)
    if (fa==0 || 
	ip[forceatoms[fa-3]].disres.label != ip[forceatoms[fa]].disres.label)
      dd->nr++;
  dd->npr = nfa/(interaction_function[F_DISRES].nratoms+1);
  snew(dd->rt,dd->npr);
  snew(dd->rav,dd->npr);
  /* Allocate Rt_6 and Rav_6 consecutively in memory so they can be
   * averaged over the processors in one call (in calc_disre_R_6)
   */
  snew(dd->Rt_6,2*dd->nr);
  dd->Rav_6 = &(dd->Rt_6[dd->nr]);
  if (mcr)
    snew(dd->Rtl_6,dd->nr);
  else
    dd->Rtl_6 = dd->Rt_6;
  
  if (dd->npr > 0) {
    fprintf(fplog,"There are %d distance restraints involving %d atom pairs\n",
	    dd->nr,dd->npr);
    if (mcr)
      check_multi_int(fplog,mcr,fcd->disres.nr,
		      "the number of distance restraints");
    please_cite(fplog,"Tropp80a");
    please_cite(fplog,"Torda89a");
  }
}

void calc_disres_R_6(const t_commrec *mcr,
		     int nfa,const t_iatom forceatoms[],const t_iparams ip[],
		     const rvec x[],const t_pbc *pbc,t_fcdata *fcd)
{
  atom_id     ai,aj;
  int         fa,res,i,pair,ki,kj,m;
  int         type,label;
  rvec        dx;
  real        rt_1,rt_3,rt2,rav_3,*rt,*rav,*Rtl_6,*Rt_6,*Rav_6;
  ivec        it,jt,dt;
  t_disresdata *dd;
  real        ETerm,ETerm1,cf1,cf2,invn=0;


  dd = &(fcd->disres);
  ETerm        = dd->ETerm;
  ETerm1       = dd->ETerm1;
  rt           = dd->rt;
  rav          = dd->rav;
  Rtl_6        = dd->Rtl_6;
  Rt_6         = dd->Rt_6;
  Rav_6        = dd->Rav_6;

  /* scaling factor to smoothly turn on the restraint forces *
   * when using time averaging                               */
  dd->exp_min_t_tau *= ETerm;

  cf1 = dd->exp_min_t_tau;
  cf2 = 1.0/(1.0 - dd->exp_min_t_tau);
  
  if (mcr)
    invn = 1.0/mcr->nnodes;

  /* 'loop' over all atom pairs (pair_nr=fa/3) involved in restraints, *
   * the total number of atoms pairs is nfa/3                          */
  res = 0;
  fa  = 0;
  while (fa < nfa) {
    type  = forceatoms[fa];
    label = ip[type].disres.label;

    Rav_6[res] = 0.0;
    Rt_6[res]  = 0.0;
  
    /* Loop over the atom pairs of 'this' restraint */
    while (fa<nfa && ip[forceatoms[fa]].disres.label==label) {
      pair = fa/3;
      ai   = forceatoms[fa+1];
      aj   = forceatoms[fa+2];

      //printf("DISRE: pair: %6d  ai: %6d aj: %6d\n",pair,ai,aj);

      if (pbc)
	pbc_dx(pbc,x[ai],x[aj],dx);
      else
	rvec_sub(x[ai],x[aj],dx);
      rt2  = iprod(dx,dx);
      rt_1 = invsqrt(rt2);
      rt_3 = rt_1*rt_1*rt_1;

      rt[pair]  = sqrt(rt2);
      rav_3     = cf2*((ETerm - cf1)*rav[pair] + ETerm1*rt_3);
      rav[pair] = rav_3;
      
      Rt_6[res]  += rt_3*rt_3;
      Rav_6[res] += rav_3*rav_3;

      fa += 3;
    }
    if (mcr) {
      Rtl_6[res]  = Rt_6[res];
      Rt_6[res]  *= invn;
      Rav_6[res] *= invn;

      ////fprintf(stderr,"node: %d res: %d pair: %d Rt_6 %7.3f dd->Rt_6 %7.3f\n",mcr->nodeid,res,pair,Rt_6[res],dd->Rt_6[res]);
      // Rt_6 and dd->Rt_6 were the same fol all values

    }
    res++;
  }

//  printf("DISRE: Rt_6 %7.3f ==> ",Rt_6[0]);  

  if ((mcr) && PAR(mcr))
    gmx_sum(2*dd->nr,Rt_6,mcr);

//   printf("SUMMED: Rt_6 %7.3f\n",Rt_6[0]);  

}

real ta_disres(int nfa,const t_iatom forceatoms[],const t_iparams ip[],
	       const rvec x[],rvec f[],rvec fshift[],
	       const t_pbc *pbc,const t_graph *g,
	       real lambda,real *dvdlambda,
	       const t_mdatoms *md,t_fcdata *fcd)
{
  const real sixth=1.0/6.0;
  const real seven_three=7.0/3.0;
  
  atom_id     ai,aj;
  int         fa,res,npairs,p,pair,ki=CENTRAL,m;
  int         type,label;
  rvec        dx;
  real        weight_rt_1;
  real        smooth_fc,Rt,Rav,rt2,*Rtl_6,*Rt_6,*Rav_6;
  real        k0,f_scal=0,fmax_scal,fk_scal,fij;
  real        tav_viol,instant_viol,mixed_viol,violtot,vtot;
  real        tav_viol_Rav7,instant_viol_Rav7;
  real        up1,up2,low;
  bool        bConservative,bMixed,bViolation;
  ivec        it,jt,dt;
  t_disresdata *dd;
  int         dr_weighting;
  bool        dr_bMixed;
  real        dr_fc;

  dd = &(fcd->disres);
  dr_weighting = dd->dr_weighting;
  dr_bMixed    = dd->dr_bMixed;
  dr_fc        = dd->dr_fc;
  Rtl_6        = dd->Rtl_6;
  Rt_6         = dd->Rt_6;
  Rav_6        = dd->Rav_6;

  tav_viol=instant_viol=mixed_viol=tav_viol_Rav7=instant_viol_Rav7=0;
  /* scaling factor to smoothly turn on the restraint forces *
   * when using time averaging                               */
  smooth_fc = dr_fc * (1.0 - dd->exp_min_t_tau); 
  
  violtot = 0;
  vtot    = 0;
  
  /* 'loop' over all atom pairs (pair_nr=fa/3) involved in restraints, *
   * the total number of atoms pairs is nfa/3                          */
  res  = 0;
  fa   = 0;
  while (fa < nfa) {
    type  = forceatoms[fa];
    /* Take action depending on restraint, calculate scalar force */
    label = ip[type].disres.label;
    up1   = ip[type].disres.up1;
    up2   = ip[type].disres.up2;
    low   = ip[type].disres.low;
    k0    = smooth_fc*ip[type].disres.kfac;

    /* Count the number of pairs in this restraint */
    npairs=1;
    while(fa + 3*npairs < nfa &&
	  ip[forceatoms[fa + 3*npairs]].disres.label == label)
      npairs++;

    /* save some flops when there is only one pair */
    if (ip[type].disres.type != 2) {
      bConservative = (dr_weighting == edrwConservative) && (npairs>1);
      bMixed        = dr_bMixed;
      Rt  = pow(Rt_6[res],-sixth);
      Rav = pow(Rav_6[res],-sixth);
    } else {
      /* When rtype=2 use the instantaneous not ensemble avereged distance */
      bConservative = npairs>1;
      bMixed        = FALSE;
      Rt  = pow(Rtl_6[res],-sixth);
      Rav = Rt;
    }
    
    if (Rav > up1) {
      bViolation = TRUE;
      tav_viol = Rav - up1;
    } else if (Rav < low) {
      bViolation = TRUE;
      tav_viol = Rav - low;
    } else
      bViolation = FALSE;

    if (bViolation) {
      /* NOTE: there is no real potential when time averaging is applied */
      vtot += 0.5*k0*sqr(tav_viol);
      if (1/vtot == 0)
	printf("vtot is inf: %f\n",vtot);
      if (!bMixed) {
	f_scal   = -k0*tav_viol;
	violtot += fabs(tav_viol);
      } else {
	if (Rt > up1) {
	  if (tav_viol > 0)
	    instant_viol = Rt - up1;
	  else
	    bViolation = FALSE;
	} else if (Rt < low) {
	  if (tav_viol < 0)
	    instant_viol = Rt - low;
	  else
	    bViolation = FALSE;
	} else
	  bViolation = FALSE;
	if (bViolation) {
	  mixed_viol = sqrt(tav_viol*instant_viol);
	  f_scal     = -k0*mixed_viol;
	  violtot   += mixed_viol;
	}
      }
    }
    /* save some flops and don't use variables when they are not calculated */
    if (bViolation) {
      fmax_scal = -k0*(up2-up1);
      /* Correct the force for the number of restraints */
      if (bConservative) {
	f_scal  = max(f_scal,fmax_scal);
	if (!bMixed)
	  f_scal *= Rav/Rav_6[res];
	else {
	  f_scal /= 2*mixed_viol;
	  tav_viol_Rav7     = tav_viol*Rav/Rav_6[res];
	  instant_viol_Rav7 = instant_viol*Rt/Rt_6[res];
	}
      } else {
	f_scal /= (real)npairs;
	f_scal  = max(f_scal,fmax_scal);
      }    
      
      /* Exert the force ... */
      
      /* Loop over the atom pairs of 'this' restraint */
      for(p=0; p<npairs; p++) {
	pair = fa/3;
	ai   = forceatoms[fa+1];
	aj   = forceatoms[fa+2];

	if (pbc) 
	  ki = pbc_dx(pbc,x[ai],x[aj],dx);
	else
	  rvec_sub(x[ai],x[aj],dx);
	rt2 = iprod(dx,dx);
	
	weight_rt_1 = invsqrt(rt2);
	
	if (bConservative) {
	  if (!dr_bMixed)
	    weight_rt_1 *= pow(dd->rav[pair],seven_three);
	  else
	    weight_rt_1 *= tav_viol_Rav7*pow(dd->rav[pair],seven_three)+
	    instant_viol_Rav7*pow(dd->rt[pair],-7);
	}
	
	fk_scal  = f_scal*weight_rt_1;

	if (g) {
	  ivec_sub(SHIFT_IVEC(g,ai),SHIFT_IVEC(g,aj),dt);
	  ki=IVEC2IS(dt);
	}

	for(m=0; m<DIM; m++) {
	  fij            = fk_scal*dx[m];
	  
	  f[ai][m]      += fij;
	  f[aj][m]      -= fij;
	  fshift[ki][m] += fij;
	  fshift[CENTRAL][m] -= fij;
	}
	fa += 3;
      }
    } else
      fa += 3*npairs;
    res++;
  }
  
  dd->sumviol = violtot;

  /* Return energy */
  return vtot;
}

/* test S2 calculation for an ensemble simulation by Szpari */

void init_s2res (FILE *fplog,int nfa,const t_iatom forceatoms[], 
//void init_s2res (FILE *fplog,int nfa,const t_iatom forceatoms[], t_topology *top,
		 const t_iparams ip[],rvec xref[],const t_mdatoms *md,
		 const t_inputrec *ir,const t_commrec *mcr,t_fcdata *fcd)
{
    if ((mcr) && PAR(mcr) && (mcr->nnodes > 1)){

       t_s2resdata *sd;
       sd=&(fcd->s2res);
       sd->nr=nfa/3;
       sd->alpha=abs(ir->s2res_fc);
       sd->do_fit = (sd->alpha == ir->s2res_fc);
       //fprintf(stderr,"alpha= %7.3f  s2res_fc = %7.3f\n",sd->alpha, ir->s2res_fc);
       snew(sd->ijvector,sd->nr);
       snew(sd->s2calc,sd->nr);
       snew(sd->s2diff,sd->nr);
       snew(sd->s2exp,sd->nr);
       snew(sd->s2err,sd->nr);
       snew(sd->rijmin4,sd->nr);
       snew(sd->x2,sd->nr);
       snew(sd->y2,sd->nr);
       snew(sd->z2,sd->nr);
       snew(sd->xy,sd->nr);
       snew(sd->xz,sd->nr);
       snew(sd->yz,sd->nr);
       sd->rhomin=-1.0;
       sd->M=mcr->nnodes;
       sd->s2reout=ir->nsts2reout;

       /* Reading in experimental values here */
       int type, fa=0, pair=0;
       while(fa < nfa){
	 type=forceatoms[fa];
	 sd->s2exp[pair]    = ip[type].s2res.S2;
	 sd->s2err[pair]    = ip[type].s2res.S2err;
         
//// A bit of test for other purposes
////         fprintf (stderr, "atom: %d, atype: %s, resnum: %d\n",forceatoms[fa+1], (*top->atoms.atomname[forceatoms[fa+1]]),top->atoms.atom[forceatoms[fa+1]].resnr);fflush(stderr);
         
         fa+=3;
	 pair++;
       }
       //exit;

       /* stuff from orires.c */
       int i,j, d;

       /* Bothering with this all only if fit needed */ 
       if (sd->do_fit){

       real mtot;
       rvec com;
       sd->nref = 0;
        for(i=0; i<md->nr; i++)
           if (md->cS2RF[i] == 0)
             sd->nref++;
        snew(sd->mref,sd->nref);
        snew(sd->xref,sd->nref);
        snew(sd->xtmp,sd->nref);

      /* Determine the reference structure on the master node.
      * Copy it to the other nodes after checking multi compatibility,
      * so we are sure the subsystems match before copying.
      */
      clear_rvec(com);
      mtot = 0.0;
      j = 0;
      for(i=0; i<md->nr; i++) {
        if (md->cS2RF[i] == 0) {
          sd->mref[j] = md->massT[i];
          if (mcr==NULL || MASTER(mcr)) {
	     copy_rvec(xref[i],sd->xref[j]);
	     for(d=0; d<DIM; d++)
	        com[d] += sd->mref[j]*xref[i][d];
	  }
          mtot += sd->mref[j];
	  j++;
	}
      }
      sd->invmref = 1.0/mtot;
      svmul(sd->invmref,com,com);
      if (mcr==NULL || MASTER(mcr))
         for(j=0; j<sd->nref; j++)
           rvec_dec(sd->xref[j],com);
  
      fprintf(fplog,"  the fit group for S2 restraints consists of %d atoms and has total mass %g\n",
	  sd->nref,mtot);


      check_multi_int(fplog,mcr,sd->nr,
		    "the number of S2 restraints");
      check_multi_int(fplog,mcr,sd->nref,
		    "the number of fit atoms for S2 restraining");
      check_multi_int(fplog,mcr,ir->nsteps,"nsteps");
      /* Copy the reference coordinates from the master to the other nodes */
      gmx_sum(DIM*sd->nref,sd->xref[0],mcr);
      }/* if do_fit*/
       
      please_cite(fplog,"Best04");

    } else if (nfa > 0){
	gmx_fatal(FARGS,"S2 calculation available only in parallel mode with at least 2 replicas!");    }
}

void calc_s2 (const t_commrec *mcr,
	      int nfa,const t_iatom forceatoms[],const t_iparams ip[],const t_mdatoms *md,
		     const rvec x[],const t_pbc *pbc,t_fcdata *fcd, int step, FILE *fplog)
{


    t_s2resdata *sd;
    sd=&(fcd->s2res);
    sd->nr=nfa/3;
    atom_id ai,aj;
    rvec dx, fx; /* Original and fitted interatomic vectors */
    rvec *ijv;
    ijv=sd->ijvector;
    int testatom1=18,testatom2=25;
    int type, m, n, d;
    real *s2calc = sd->s2calc;
    real S2;
    real invn;

    real rijmin=0, bondlen, actual;
    real *rijmin4=sd->rijmin4;

    int fa=0, pair=0;
    int ndn;
    tensor R;

    real *x2=sd->x2;
    real *y2=sd->y2;
    real *z2=sd->z2;
    real *xy=sd->xy;
    real *xz=sd->xz;
    real *yz=sd->yz;

    /* from orires.c */
    int nref = sd->nref;
    real *mref = sd->mref;
    rvec *xref = sd->xref;
    rvec *xtmp = sd->xtmp;
    rvec com;
    int i,j;


    if (sd->do_fit){
      clear_rvec(com);
      j=0;
      for(i=0; i<md->nr; i++)
	if (md->cS2RF[i] == 0) {
	  copy_rvec(x[i],xtmp[j]);
	  for(d=0; d<DIM; d++)
	    com[d] += mref[j]*xref[j][d];
	  j++;
	}
      svmul(sd->invmref,com,com);
      for(j=0; j<nref; j++)
	rvec_dec(xtmp[j],com);

      calc_fit_R(nref,mref,xref,xtmp,R);
    }

    if (mcr){
	invn=1.0/mcr->nnodes;
        ndn=mcr->nodeid;

        // Calculating vector components for each i,j pair
	while (fa<nfa) {
	    pair = fa/3;
	    type = forceatoms[fa];
	    ai   = forceatoms[fa+1];
	    aj   = forceatoms[fa+2];

	    //fprintf(stderr,"s2res pair: %d (%d - %d) fa: %d\n",pair,ai,aj,fa);

	    if (pbc)
		pbc_dx(pbc,x[aj],x[ai],dx);
	    else
		rvec_sub(x[aj],x[ai],dx);

      	    /* bondlen: -1 => bond length at the first step is used throughout
                         0 => bond length is recaclculated every step
                        >0 => supplied value used
	     */

	   bondlen    = ip[type].s2res.bondlen;
	   actual=norm(dx);

	   if ((bondlen == -1.0) && (sd->rhomin == -1)){
		rijmin=norm(dx);
		rijmin4[pair]=rijmin*rijmin*rijmin*rijmin;
	   } else if (bondlen == 0.0){
		rijmin=norm(dx);
		rijmin4[pair]=rijmin*rijmin*rijmin*rijmin;
	   } else if (bondlen > 0.0){
	        rijmin=bondlen;
		rijmin4[pair]=rijmin*rijmin*rijmin*rijmin;
	   }

	   //fprintf(stderr,"pair: %d bondlen: %7.3f rijmin %7.3f, actual %7.3f, fijmin4 %7.6f\n",pair,bondlen,rijmin,actual,rijmin4[pair]);

     
           /* Transforming the bond vector (half the time than the two bond vectors :-) */
	   if (sd->do_fit){
	     mvmul(R,dx,fx);
	     copy_mat(R,sd->R);
	   } else {
           /* Go on with fx even if there is no fit -> so copy dx to it */
#ifdef GMX_DOUBLE
	     copy_dvec(dx,fx);  
#else
	     copy_rvec(dx,fx);  
#endif
	   }

            /* Storing the transformed vector coordinates as force will be calculated in the transformed frame first */
#ifdef GMX_DOUBLE
	    copy_dvec(fx,ijv[pair]);  
#else
	    copy_rvec(fx,ijv[pair]);  
#endif

	    x2[pair]=fx[0]*fx[0]*invn;
	    y2[pair]=fx[1]*fx[1]*invn;
	    z2[pair]=fx[2]*fx[2]*invn;
	    xy[pair]=fx[0]*fx[1]*invn;
	    xz[pair]=fx[0]*fx[2]*invn;
	    yz[pair]=fx[1]*fx[2]*invn;

	    fa+=3;
	}    
    } else fprintf(stderr,"S2 calculation available in ensemble mode only\n");


    // Calculating simulated S2 values
    // Aim: to simlpify averaging over procs with a memory trick like for R_av
    if ((mcr) && PAR(mcr)){
	gmx_sum(sd->nr,x2,mcr);
	gmx_sum(sd->nr,y2,mcr);
	gmx_sum(sd->nr,z2,mcr);
	gmx_sum(sd->nr,xy,mcr);
	gmx_sum(sd->nr,xz,mcr);
	gmx_sum(sd->nr,yz,mcr);

	fa=0;
	pair=0;
	while (fa < nfa){
	  type=forceatoms[fa];
	    s2calc[pair]=(1.5/rijmin4[pair])*(x2[pair]*x2[pair] + y2[pair]*y2[pair] + z2[pair]*z2[pair] + 2*xy[pair]*xy[pair] + 2*xz[pair]*xz[pair] + 2*yz[pair]*yz[pair])-0.5;

	    S2    = ip[type].s2res.S2;
            if ((do_per_step(step,sd->s2reout) == 1) && MASTER(mcr))
	      fprintf (fplog,"S2RES: step %d : S2 for pair %d (atoms %d and %d) = %7.3f EXP: %7.3f\n", step,pair,forceatoms[fa+1],forceatoms[fa+2],s2calc[pair],S2);
	    
	    fa += 3;
	    pair++;
	}
        if ((do_per_step(step,sd->s2reout) == 1) && MASTER(mcr))
	  fprintf(fplog,"\n");
    }
}

real ta_s2res(int nfa,const t_iatom forceatoms[],const t_iparams ip[],
	       const rvec x[],rvec f[],rvec fshift[],
	       const t_pbc *pbc,const t_graph *g,
	       real lambda,real *dvdlambda,
	       const t_mdatoms *md,t_fcdata *fcd)
{


//    fprintf(stderr,"Maxwell's silver hammer\n");

    t_s2resdata *sd;
    sd=&(fcd->s2res);
    sd->rho=0;

    real rho=sd->rho;
    real rhomin=sd->rhomin;

    //sd->rhomin=-0.01;

    real *S2=sd->s2exp; 
    real *S2err=sd->s2err;

    int type, ai,aj;
    real vtot=0.0; //fij=0.0;

    rvec rfij, fij;

    int ki=CENTRAL;
    rvec dx;
    ivec dt;

    real *x2=sd->x2;
    real *y2=sd->y2;
    real *z2=sd->z2;
    real *xy=sd->xy;
    real *xz=sd->xz;
    real *yz=sd->yz;

    rvec *ijv=sd->ijvector;
    real *rijmin4=sd->rijmin4;

    real *s2diff = sd->s2diff;


    // alpha is a global thing, not a per restraint measure
    //real alpha=500.00;
    real alpha=sd->alpha;
    //fprintf (stderr,"ALPHA=%7.3f\n",alpha);fflush(stderr);

    real *s2calc=sd->s2calc;

    //snew(S2,sd->nr);
    //snew(S2err,sd->nr);

    int M=sd->M; //conforming the notation in Best & Vendruscolo
    int Ne=sd->nr;
    int pair, fa, m;

    fa=0;
    pair=0;
    rho=0.0;
    tensor invR;

    if (sd->do_fit){    
      /* Matrix to rotate force components back */
      m_inv(sd->R,invR);
    }

    clear_rvec(rfij);
    clear_rvec(fij);

    while(fa < nfa){
      //type=forceatoms[fa];

	//S2[pair]    = ip[type].s2res.S2;
	//S2err[pair] = ip[type].s2res.S2err;


	//fprintf(stderr,"pair: %d  S2: %7.3f (%7.3f) S2calc: %7.3f\n",pair,S2[pair],S2err[pair],s2calc[pair]);fflush(stderr);
        /* Allowing for errors also (errors can be set to zero, in this case the stuff below is identical
           to the original treatment which is already half-harmonic!) */
	////	if ((s2calc[pair] < S2[pair]-S2err[pair]) || (s2calc[pair] > S2[pair]+S2err[pair]))  
	if (s2calc[pair] < S2[pair]-S2err[pair]){
	  s2diff[pair]=s2calc[pair]-(S2[pair]-S2err[pair]);
	} else if (s2calc[pair] > S2[pair]+S2err[pair]){
	  s2diff[pair]=s2calc[pair]-(S2[pair]+S2err[pair]);
	} else{
	  s2diff[pair]=0.0;
	}

	//rho+=(S2[pair]-s2calc[pair])*(S2[pair]-s2calc[pair]);
	rho+=(s2diff[pair]*s2diff[pair]);

	fa+=3;
	pair++;
    }

    rho/=Ne;
    sd->rho=rho;

    //if (MASTER(mcr))
     //fprintf(stderr,"Rho = %7.5f, Rhomin = %7.5f, Ne=%d\n",rho,rhomin,Ne);fflush(stderr);

    // No rhomin for the 1st calculation, so skip the main part
    if (rhomin < 0.0){
        sd->rhomin=rho;
	vtot=0.0;
    } else if (rho > rhomin){
	vtot=(alpha*M/2)*(rho-rhomin)*(rho-rhomin);


	fa=0;
	pair=0;
	while(fa < nfa){
	  // Calculating force only whne there is violation for that restraint
	  //if ((s2calc[pair] < S2[pair]-S2err[pair]) || (s2calc[pair] > S2[pair]+S2err[pair])){
	    ai   = forceatoms[fa+1];
	    aj   = forceatoms[fa+2];

	    if (pbc)
		ki=pbc_dx(pbc,x[ai],x[aj],dx);

	    if (g) {
		ivec_sub(SHIFT_IVEC(g,ai),SHIFT_IVEC(g,aj),dt);
		ki=IVEC2IS(dt);
	    }
  
            ////fprintf(stderr,"pair: %d x2 %7.3f\n",pair,x2[pair]);

            /* Calculating forces for the transformed molecule first then rotating the force components back */
            //X
	    rfij[0]=(12*alpha/Ne)*(rho-rhomin)*(s2diff[pair]/rijmin4[pair])*(x2[pair]*ijv[pair][0]+xy[pair]*ijv[pair][1]+xz[pair]*ijv[pair][2]);
            //Y
	    rfij[1]=(12*alpha/Ne)*(rho-rhomin)*(s2diff[pair]/rijmin4[pair])*(y2[pair]*ijv[pair][1]+yz[pair]*ijv[pair][2]+xy[pair]*ijv[pair][0]);
            //Z
	    rfij[2]=(12*alpha/Ne)*(rho-rhomin)*(s2diff[pair]/rijmin4[pair])*(z2[pair]*ijv[pair][2]+xz[pair]*ijv[pair][0]+yz[pair]*ijv[pair][1]);


            /* If there was fitting, rotate force components back, otherwise simply copy them */
            if (sd->do_fit){
	      mvmul(invR,rfij,fij);
	    } else {
#ifdef GMX_DOUBLE
	      copy_dvec(rfij,fij);  
#else
	      copy_rvec(rfij,fij);  
#endif
	    }            

	    //fprintf(stderr,"FORCE! pair %d F %7.3f %7.3f %7.3f\n",pair,fij[0],fij[1],fij[2]);fflush(stderr);


	    for(m=0; m<DIM; m++) {
	      f[ai][m]      += fij[m];
	      f[aj][m]      -= fij[m];
	      fshift[ki][m] += fij[m];
	      fshift[CENTRAL][m] -= fij[m];
	    }

	    fa+=3;
	    pair++;

	}


        //sd->rhomin=sd->rho;


    }
    if (rho < rhomin)
	sd->rhomin=rho;
    
    return vtot;

}


/* test HX protection factor calculation by Szpari */

void init_hxres (FILE *fplog,int nfa,const t_iatom forceatoms[], t_topology *top,
		 const t_iparams ip[],rvec xref[],const t_mdatoms *md,
		 const t_inputrec *ir,const t_commrec *mcr,t_fcdata *fcd)
{

  t_hxresdata *hd;
  hd=&(fcd->hxres);
  hd->nr=nfa/3;
  hd->fc_hx=abs(ir->hxres_fc);
  // Will do ensemble averaging if sign matches
  hd->bEnsemble = (hd->fc_hx == ir->hxres_fc);
  hd->bNitAcc = ir->hxres_bnitacc;


  hd->cseqdist = ir->hxres_minseqdist;
  //fprintf(stderr,"cseqdist: %d\n",hd->cseqdist);
  hd->ccut = ir->hxres_ccut;
  hd->hcut = ir->hxres_hcut;
  hd->beta_c = ir->hxres_betac;
  hd->beta_h = ir->hxres_betah;
  hd->fc_inc_steps = ir->hxres_fc_steps;
  hd->rhomin=-1.0;
  hd->hxreout=ir->nsthxreout;

  if (mcr)   hd->M=mcr->nnodes;
  else hd->M = 1;

  snew(hd->Pexp,hd->nr);
  snew(hd->Perr,hd->nr);
  snew(hd->Psim,hd->nr);

  snew(hd->contacts,hd->nr);
  snew(hd->hbonds,hd->nr);

  snew(hd->Nresnum,hd->nr);



   /* Reading in experimental values here */

   int type, fa=0, pair=0;
   while(fa < nfa){
       type=forceatoms[fa];

       // We work with ln(P) but expect P n the topology file!
       hd->Pexp[pair]    = log(ip[type].hxres.P);
       // According to the error propagation rules I found:
       hd->Perr[pair]    = ip[type].hxres.Perr/ip[type].hxres.P;
  
       hd->Nresnum[pair]=top->atoms.atom[forceatoms[fa+1]].resnr;
       fa+=3;
       pair++;
   }
    
  snew(hd->atomid_cal,top->atoms.nr);
  snew(hd->resnum_cal,top->atoms.nr);
  snew(hd->hbacc_cal,top->atoms.nr);

  int i,j=0;

  if (nfa > 0){
   // building list of aotms used for contact calculations
   // presently this means heavy atoms
   for (i=0; (i<top->atoms.nr); i++) {
      // heavy atom
      if ((*top->atoms.atomname[i])[0] != 'H'){
	  hd->atomid_cal[j] = i;
	  hd->resnum_cal[j] = top->atoms.atom[i].resnr;
          // H-bond acceptor?
	  hd->hbacc_cal[j] = (((*top->atoms.atomname[i])[0] == 'O') || (hd->bNitAcc && ((*top->atoms.atomname[i])[0] == 'N')));
	  j++;
      }
   }//i

   // wonder if this will work
   fprintf(stderr,"HX restraints: Readjusting memory allocation from %d to %d\n",i,j);
   hd->calnum=j;
   srenew(hd->atomid_cal,j);
   srenew(hd->resnum_cal,j);
   srenew(hd->hbacc_cal,j);

   please_cite(fplog,"Gsponer06");

   if (!(mcr) && (hd->bEnsemble)){
     gmx_fatal(FARGS,"Cannot do ensemble HX restraint calculation for 1 replica!");
   }
  }

}


void calc_hx (const t_commrec *mcr,
	      int nfa,const t_iatom forceatoms[],const t_iparams ip[],const t_mdatoms *md,
		     const rvec x[],const t_pbc *pbc,t_fcdata *fcd, int step, FILE *fplog)
{

    t_hxresdata *hd;
    hd=&(fcd->hxres);
    hd->rho=0;

    real rho=hd->rho;
    real rhomin=hd->rhomin;

    real ccut=hd->ccut;
    real hcut=hd->hcut;

    real beta_c=hd->beta_c;
    real beta_h=hd->beta_h;
    int cseqdist =  hd->cseqdist;

    // making the sigmoid curve more steplike
    // this seems to make H-bonds really 1/0-like (but only time will tell...)
    // currently hardcoded, should be modified in ta_hxres also! 
    // same as "beta" in the Vendruscolo-group implementation
    real steplike=500.0;

    real *Pexp=hd->Pexp;
    real *Perr=hd->Perr;
    real *Psim=hd->Psim;
    int Nresnum;
    int Natomnum;
    int j;
    atom_id ai,aj;

    rvec dx;
    real r;


    int type, fa=0, pair=0;
    while(fa < nfa){
	type=forceatoms[fa];
	Natomnum=forceatoms[fa+1];
	Nresnum=hd->Nresnum[pair];

	hd->hbonds[pair]=0;
	hd->contacts[pair]=0;

        // H atom
	// ai   = forceatoms[fa+2];
        // N atom
	// aj   = forceatoms[fa+1];

	//fprintf(stderr,"hxres pair: %d (%d - %d) fa: %d\n",pair,ai,aj,fa);

        // counting contacts and H-bonds (note the roles of ai and aj for N and H!)
	for  (j=0; j<hd->calnum; j++){
            // heavy atom contacts form amide N (as stated in the suppl. text of Gsponer 2006)
	    if (abs(hd->resnum_cal[j]-Nresnum) >= cseqdist){
		ai   = hd->atomid_cal[j];
		aj   = forceatoms[fa+1];
		if (pbc)
		    pbc_dx(pbc,x[aj],x[ai],dx);
	        else
		    rvec_sub(x[aj],x[ai],dx);

		r=norm(dx);

		//fprintf(stderr,"pair %d: atoms %d (rn: %d) & %d (rn: %d) : r = %7.3f ccut: %7.3f (cd: %d)\n",pair,forceatoms[fa+2],Nresnum,hd->atomid_cal[j],hd->resnum_cal[j],r,ccut,cseqdist);
		hd->contacts[pair]+=1/(1+exp(steplike*(r-ccut)));
		//if (r <= ccut) hd->contacts[pair]+=1.0; // beta_c;
	    }
            // H-bonds
            // Excluding covalently bound N-atom but allowing all other acceptors
	    if ((Natomnum != hd->atomid_cal[j]) && (hd->hbacc_cal[j])){
		ai   = forceatoms[fa+2];
		aj   = hd->atomid_cal[j];
		if (pbc)
		    pbc_dx(pbc,x[aj],x[ai],dx);
	        else
		    rvec_sub(x[aj],x[ai],dx);

		r=norm(dx);
                //fprintf(stderr,"pair %d: atoms %d (rn: %d) & %d (rn: %d) : r = %7.3f hcut: %7.3f\n",pair,forceatoms[fa+2],Nresnum,hd->atomid_cal[j],hd->resnum_cal[j],r,hcut);
		hd->hbonds[pair]+=1/(1+exp(steplike*(r-hcut)));
		//if (r <= hcut) hd->hbonds[pair]+=1.0; //beta_h;
	    }

	}

	//fprintf(stderr,"HX restraints: pair %d : contacts: %7.3f : H-bonds: %7.3f\n",pair,hd->contacts[pair],hd->hbonds[pair]),

	// Calculating ln(P) as in Gsponer et al., 2006
	Psim[pair]=beta_c*hd->contacts[pair]+beta_h*hd->hbonds[pair];

	fa+=3;
	pair++;
    }

    // Calculating ensemble average 
    if ((mcr) && (PAR(mcr)) && (hd->bEnsemble)){
	real invn=1.0/mcr->nnodes;
	for (pair=0; pair<hd->nr; pair++) Psim[pair]*=invn;
	gmx_sum(hd->nr,Psim,mcr);
    }


    // Writing HX data to log file
    bool writelog=(do_per_step(step,hd->hxreout)==1);
    if (mcr){
	if ((hd->bEnsemble) && (!MASTER(mcr))) writelog=FALSE;
    }


//    if ((do_per_step(step,hd->hxreout) == 1) && (MASTER(mcr)) || (hd->M==1)){
    if (writelog){
	fa=0;pair=0;
	while(fa < nfa){
	    //type=forceatoms[fa],
	    //fprintf (fplog,"HXRES: step %d : ln(P) for pair %d (atoms %d and %d) = %7.3f EXP: %7.3f\n", step,pair,forceatoms[fa+1],forceatoms[fa+2],Psim[pair],Pexp[pair]);
	    fa+=3;
	    pair++;
	}
	fprintf(fplog,"\n");
    }// if write log


}// calc_hx

real ta_hxres(int nfa,const t_iatom forceatoms[],const t_iparams ip[],
	       const rvec x[],rvec f[],rvec fshift[],
	       const t_pbc *pbc,const t_graph *g,
	       real lambda,real *dvdlambda,
	       const t_mdatoms *md,t_fcdata *fcd)
{

    t_hxresdata *hd;
    hd=&(fcd->hxres);
    hd->rho=0;

    int M=hd->M;
    real fc_hx=hd->fc_hx;

    real rho=hd->rho;
    real rhomin=hd->rhomin;

    real ccut=hd->ccut;
    real hcut=hd->hcut;

    real beta_c=hd->beta_c;
    real beta_h=hd->beta_h;
    int cseqdist =  hd->cseqdist;

    // making the sigmoid curve more steplike
    // this seems to make H-bonds really 1/0-like (but only time will tell...)
    real steplike=500.0;

    real *Pexp=hd->Pexp;
    real *Perr=hd->Perr;
    real *Psim=hd->Psim;
    int Nresnum;
    int Natomnum;
    int j,m;
    atom_id ai,aj;

    int fa, pair, type;

    rvec dx;
    ivec dt;
    real r;

    real vtot=0.0; //fij=0.0;

    rvec fij;
    int ki=CENTRAL;

    // Calculating rho
    int i;
    hd->rho=0.0;
    for (i=0; i<hd->nr; i++){
	hd->rho+=(Psim[i]-Pexp[i])*(Psim[i]-Pexp[i]);
    }
    hd->rho/=hd->nr;
    rho=hd->rho;

    //fprintf(stderr,"HXRES rho: %7.3f  rhomin: %7.3f\n",hd->rho,hd->rhomin);

    // Checking rho
    if (rhomin < 0.0){
        hd->rhomin=rho;
	vtot=0.0;
    } else if (rho > rhomin){
	vtot=(fc_hx*M/2)*(rho-rhomin)*(rho-rhomin);

        // using similar variable names as the Vendruscolo group
	real bux=-2*fc_hx*(rho-rhomin)/hd->nr;
	real cux,a,b,c;

	fa=0;pair=0;
	while(fa < nfa){
	    type=forceatoms[fa];
	    Natomnum=forceatoms[fa+1];
	    Nresnum=hd->Nresnum[pair];

	    cux=Psim[pair]-Pexp[pair];

            // H atom
	    // ai   = forceatoms[fa+2];
            // N atom
	    // aj   = forceatoms[fa+1];

            // forces for contacts and H-bonds (note the roles of ai and aj for N and H!)
  	    for  (j=0; j<hd->calnum; j++){
            // heavy atom contacts form amide N (as stated in the suppl. text of Gsponer 2006)
	     if (abs(hd->resnum_cal[j]-Nresnum) >= cseqdist){
		ai   = hd->atomid_cal[j];
		aj   = forceatoms[fa+1];
		if (pbc)
		    pbc_dx(pbc,x[aj],x[ai],dx);
	        else
		    rvec_sub(x[aj],x[ai],dx);

		r=norm(dx);

		a=exp(steplike*(r-ccut));
		b=1/((1+a)*(1+a));
		c=a*b*steplike/r;


		if (pbc)
		    ki=pbc_dx(pbc,x[ai],x[aj],dx);
		if (g) {
		    ivec_sub(SHIFT_IVEC(g,ai),SHIFT_IVEC(g,aj),dt);
		    ki=IVEC2IS(dt);
		}
                // signs here should be checked & tested!!
		for(m=0; m<DIM; m++) {
		    f[ai][m]      -= c*cux*bux*dx[m];
		    f[aj][m]      += c*cux*bux*dx[m];
		    fshift[ki][m] -= c*cux*bux*dx[m];
		    fshift[CENTRAL][m] += c*cux*bux*dx[m];
		}
	     }// contacts
             // H-bonds
             // Excluding covalently bound N-atom but allowing all other acceptors
	     if ((Natomnum != hd->atomid_cal[j]) && (hd->hbacc_cal[j])){
		ai   = forceatoms[fa+2];
		aj   = hd->atomid_cal[j];
		if (pbc)
		    pbc_dx(pbc,x[aj],x[ai],dx);
	        else
		    rvec_sub(x[aj],x[ai],dx);

		r=norm(dx);

		r=norm(dx);

		a=exp(steplike*(r-hcut));
		b=1/((1+a)*(1+a));
		c=a*b*steplike/r;


		if (pbc)
		    ki=pbc_dx(pbc,x[ai],x[aj],dx);
		if (g) {
		    ivec_sub(SHIFT_IVEC(g,ai),SHIFT_IVEC(g,aj),dt);
		    ki=IVEC2IS(dt);
		}
		for(m=0; m<DIM; m++) {
		    f[ai][m]      += c*cux*bux*dx[m];
		    f[aj][m]      -= c*cux*bux*dx[m];
		    fshift[ki][m] += c*cux*bux*dx[m];
		    fshift[CENTRAL][m] -= c*cux*bux*dx[m];
		}
	     }// H-bonds



	    }// for j
	    fa+=3;
	    pair++;
	}

    }// if rho > rhomin
    if (rho < rhomin)
	hd->rhomin=rho;

    return vtot;
}


/* Half-harmonic distance restraints with MUMO support */
// Time averaging will not be used, trying not to eliminate, though (may come handy)
void init_hhdisres(FILE *fplog,int nfa,const t_iatom forceatoms[],
		 const t_iparams ip[],
		 const t_inputrec *ir,const t_commrec *mcr,t_fcdata *fcd)
{
  int          fa;
  t_hhdisresdata *dd;

  dd = &(fcd->hhdisres);

  dd->dr_weighting = ir->eDisreWeighting;
  // Encoding mode with signs 
  dd->dr_Mode      = 1;

  dd->rho          = 0.0;
  dd->rhomin       = -1.0;

  if (mcr)
      dd->M            = mcr->nnodes;
  else 
      dd->M=1;



  if (ir->hhdr_fc < 0) dd->dr_Mode=2;
  if (ir->hhdr_exp < 0) dd->dr_Mode=3;
  dd->dr_fc        = abs(ir->hhdr_fc);
  dd->dr_tau       = ir->hhdr_tau;
  dd->dr_exp       = abs(ir->hhdr_exp);

  //fprintf(stderr,"Mode: %d (fc: %7.3f exp: %d)\n",dd->dr_Mode,ir->hhdr_fc,ir->hhdr_exp);fflush(stderr);

  if (dd->dr_tau == 0.0) {
    dd->dr_bMixed = FALSE;
    dd->ETerm = 0.0;
  } else {
    dd->dr_bMixed = ir->bDisreMixed;
    dd->ETerm = exp(-(ir->delta_t/ir->dr_tau));
  }
  dd->ETerm1 = 1.0 - dd->ETerm;
  dd->exp_min_t_tau = 1.0;
  
  dd->nr = 0;
  for(fa=0; fa<nfa; fa+=3)
    if (fa==0 || 
	ip[forceatoms[fa-3]].hhdisres.label != ip[forceatoms[fa]].hhdisres.label)
      dd->nr++;
  dd->npr = nfa/(interaction_function[F_HHDISRES].nratoms+1);
  snew(dd->rt,dd->npr);
  snew(dd->rav,dd->npr);

  // MUMO stuff
  snew(dd->Rt_left,dd->nr);
  snew(dd->Rt_right,dd->nr);
  snew(dd->violl,dd->nr);
  snew(dd->violr,dd->nr);

  // general half-harmonic stuff
  snew(dd->viol,dd->nr);
  snew(dd->rnij,dd->nr);

  /* Allocate Rt_6 and Rav_6 consecutively in memory so they can be
   * averaged over the processors in one call (in calc_disre_R_6)
   */

  //fprintf(stderr,"MUMO test initiated\n"),fflush(stderr);

  snew(dd->Rt_6,2*dd->nr);
  dd->Rav_6 = &(dd->Rt_6[dd->nr]);
  if (mcr)
    snew(dd->Rtl_6,dd->nr);
  else
    dd->Rtl_6 = dd->Rt_6;
  
  if (dd->npr > 0) {
    fprintf(fplog,"There are %d Half-harmonic distance restraints involving %d atom pairs\n",
	    dd->nr,dd->npr);
    if (mcr)
      check_multi_int(fplog,mcr,fcd->hhdisres.nr,
		      "the number of distance restraints");

  }

  please_cite(fplog,"Richter07");
}

void calc_hhdisres(const t_commrec *mcr,
		     int nfa,const t_iatom forceatoms[],const t_iparams ip[],
		     const rvec x[],const t_pbc *pbc,t_fcdata *fcd)
{
  atom_id     ai,aj;
  int         fa,res,i,pair,ki,kj,m;
  int         type,label;
  rvec        dx;
  real        rt_1,rt_3,rt2,rav_3,*rt,*rav,*Rtl_6,*Rt_6,*Rav_6,*Rt_left,*Rt_right,*RNij;
  ivec        it,jt,dt;
  t_hhdisresdata *dd;
  real        ETerm,ETerm1,cf1,cf2,invn=0;
  int         exp;

// MUMO!!
  double *mbuf[2];
  int u, mbufs;
// MUMO!!



  dd = &(fcd->hhdisres);
  ETerm        = dd->ETerm;
  ETerm1       = dd->ETerm1;
  rt           = dd->rt;
  rav          = dd->rav;
  Rtl_6        = dd->Rtl_6;
  Rt_6         = dd->Rt_6;
  Rt_left      = dd->Rt_left;
  Rt_right     = dd->Rt_right;
  Rav_6        = dd->Rav_6;
  exp          = dd->dr_exp;
  RNij         = dd->rnij;

  /* scaling factor to smoothly turn on the restraint forces *
   * when using time averaging                               */
  dd->exp_min_t_tau *= ETerm;

  cf1 = dd->exp_min_t_tau;
  cf2 = 1.0/(1.0 - dd->exp_min_t_tau);
  
  if (mcr)
    invn = 1.0/mcr->nnodes;

  /* 'loop' over all atom pairs (pair_nr=fa/3) involved in restraints, *
   * the total number of atoms pairs is nfa/3                          */
  res = 0;
  fa  = 0;

  real rnij_x=0;
  real rnij_y=0;
  real rnij_z=0;

  while (fa < nfa) {
    type  = forceatoms[fa];
    label = ip[type].hhdisres.label;

    Rav_6[res] = 0.0;
    Rt_6[res]  = 0.0;

    //fprintf(stderr,"fa: %d\n",fa);fflush(stderr);
  
    /* Loop over the atom pairs of 'this' restraint */

    while (fa<nfa && ip[forceatoms[fa]].hhdisres.label==label) {
      pair = fa/3;
      ai   = forceatoms[fa+1];
      aj   = forceatoms[fa+2];

      //fprintf(stderr,"DISRE: pair: %6d  ai: %6d aj: %6d\n",pair,ai,aj);fflush(stderr);

      if (pbc)
	pbc_dx(pbc,x[ai],x[aj],dx);
      else
	rvec_sub(x[ai],x[aj],dx);
      rt2  = iprod(dx,dx);
      rt_1 = invsqrt(rt2);
      rt_3 = rt_1*rt_1*rt_1;

      rt[pair]  = sqrt(rt2);
      rav_3     = cf2*((ETerm - cf1)*rav[pair] + ETerm1*rt_3);
      rav[pair] = rav_3;
      
      Rt_6[res]  += rt_3*rt_3;

      rnij_x+=dx[0]; rnij_y+=dx[1]; rnij_z+=dx[2];

      // There is some controversy about the exponent, let the user decide.
      if (exp == 3)
	  Rav_6[res] += rav_3;
      else if (exp == 6)
	  Rav_6[res] += rav_3*rav_3;

      fa += 3;
    }

    RNij[res]=sqrt(rnij_x*rnij_x+rnij_y*rnij_y+rnij_z*rnij_z);

    //fprintf(stderr,"In calc\n");fflush(stderr);

    if (mcr) {
      Rtl_6[res]  = Rt_6[res];
      Rt_6[res]  *= invn;
      Rav_6[res] *= invn;

      Rt_left[res]=Rav_6[res]/2.0;
      Rt_right[res]=Rav_6[res]/2.0;

      ////fprintf(stderr,"node: %d res: %d pair: %d Rt_6 %7.3f dd->Rt_6 %7.3f\n",mcr->nodeid,res,pair,Rt_6[res],dd->Rt_6[res]);
      // Rt_6 and dd->Rt_6 were the same fol all values

    }
    res++;
  }

//  printf("DISRE: Rt_6 %7.3f ==> ",Rt_6[0]);  

  if ((mcr) && PAR(mcr)){
    gmx_sum(2*dd->nr,Rt_6,mcr);
  }
  if ((mcr) && (dd->dr_Mode == 3)){

	snew(mbuf[0],dd->nr);
	snew(mbuf[1],dd->nr);

/* MUMO test code */
/*
	srenew(Rt_left,2);
	srenew(Rt_right,2);
	Rt_left[0]=(float)mcr->nodeid/2;
	Rt_right[0]=(float)mcr->nodeid;
	fprintf(stderr,"MUMU: Node: %d, leftavg: %7.3f, rightavg: %7.3f\n",mcr->nodeid,Rt_left[0],Rt_right[0]);fflush(stderr);
 //MUMO */

	mbufs=dd->nr*sizeof(mbuf[0][0]);

        // Left neighbor
	for (u=0; u < dd->nr; u++)
	    mbuf[0][u]=Rt_left[u];

        gmx_tx(mcr->right,mbuf[0],mbufs);
        gmx_rx(mcr->left,mbuf[1],mbufs);
	gmx_tx_wait(mcr->right);
	gmx_rx_wait(mcr->left);
	for (u=0; u < dd->nr; u++)
	    Rt_left[u]+=mbuf[1][u];

        // Right neighbor
	for (u=0; u < dd->nr; u++)
	    mbuf[0][u]=Rt_right[u];

        gmx_tx(mcr->left,mbuf[0],mbufs);
        gmx_rx(mcr->right,mbuf[1],mbufs);
	gmx_tx_wait(mcr->left);
	gmx_rx_wait(mcr->right);
	for (u=0; u < dd->nr; u++)
	    Rt_right[u]+=mbuf[1][u];

/* MUMO test code */


/*
	fprintf(stderr,"MUMO: Node: %d, leftavg: %7.3f, rightavg: %7.3f\n",mcr->nodeid,Rt_left[0],Rt_right[0]);fflush(stderr);
 //MUMO */

	sfree(mbuf[1]);
	sfree(mbuf[0]);

    }
  //}
//   printf("SUMMED: Rt_6 %7.3f\n",Rt_6[0]);  

}

real ta_hhdisres(int nfa,const t_iatom forceatoms[],const t_iparams ip[],
	       const rvec x[],rvec f[],rvec fshift[],
	       const t_pbc *pbc,const t_graph *g,
	       real lambda,real *dvdlambda,
	       const t_mdatoms *md,t_fcdata *fcd)
{
  const real sixth=1.0/6.0;
  const real seven_three=7.0/3.0;

//  fprintf(stderr,"in hh_tadisres\n");fflush(stderr);
  
  atom_id     ai,aj;
  int         fa,res,npairs,p,pair,ki=CENTRAL,m;
  int         type,label;
  rvec        dx;
  real        weight_rt_1;
  real        smooth_fc,Rt,Rav,rt2,*Rtl_6,*Rt_6,*Rav_6, *RNij;
  real        k0,f_scal=0,fmax_scal,fk_scal,fij;
  real        tav_viol,instant_viol,mixed_viol,violtot,vtot;
  real        tav_viol_Rav7,instant_viol_Rav7;
  real        up1,up2,low;
  bool        bConservative,bMixed,bViolation;
  ivec        it,jt,dt;
  t_hhdisresdata *dd;
  int         dr_weighting;
  bool        dr_bMixed;
  real        dr_fc;
  real        *viol, *violl, *violr;

  dd = &(fcd->hhdisres);
  dr_weighting = dd->dr_weighting;
  dr_bMixed    = dd->dr_bMixed;
  dr_fc        = dd->dr_fc;
  Rtl_6        = dd->Rtl_6;
  Rt_6         = dd->Rt_6;
  Rav_6        = dd->Rav_6;
  viol         = dd->viol;
  violl        = dd->violl;
  violr        = dd->violr;
  RNij         = dd->rnij;

  real oneperexp=1.0/(real)dd->dr_exp;
  real *Rt_left,*Rt_right;
  Rt_left        = dd->Rt_left;
  Rt_right        = dd->Rt_right;
  real r;

  tav_viol=instant_viol=mixed_viol=tav_viol_Rav7=instant_viol_Rav7=0;
  /* scaling factor to smoothly turn on the restraint forces *
   * when using time averaging                               */
  smooth_fc = dd->dr_fc * (1.0 - dd->exp_min_t_tau); 
  
  violtot = 0;
  vtot    = 0;

  dd->rho=0.0;

  
  /* 'loop' over all atom pairs (pair_nr=fa/3) involved in restraints, *
   * the total number of atoms pairs is nfa/3                          */

  res  = 0;
  fa   = 0;
  while (fa < nfa) {
    type  = forceatoms[fa];
    /* Take action depending on restraint, calculate scalar force */
    label = ip[type].hhdisres.label;
    up1   = ip[type].hhdisres.up1;
    up2   = ip[type].hhdisres.up2;
    low   = ip[type].hhdisres.low;
    k0    = smooth_fc*ip[type].hhdisres.kfac;

    /* Count the number of pairs in this restraint */
    npairs=1;
    while(fa + 3*npairs < nfa &&
	  ip[forceatoms[fa + 3*npairs]].hhdisres.label == label)
      npairs++;




    // MUMO-type rho calculation
    if (dd->dr_Mode == 3){


	Rt_left[res] = pow(Rt_left[res],-oneperexp);
	Rt_right[res] = pow(Rt_right[res],-oneperexp);



	violl[res]=Rt_left[res] > up1 ? Rt_left[res]-up1 : Rt_left[res] < low ? low-Rt_left[res] : 0.0;
	dd->rho += 0.5*violl[res]*violl[res];

	//fprintf(stderr,"left (%d): %7.3f up1: %7.3f low: %7.3f  viol: %7.3f\n",res,Rt_left[res],up1,low,violl[res]);fflush(stdout);

	violr[res]=Rt_right[res] > up1 ? Rt_right[res]-up1 : Rt_right[res] < low ? low-Rt_right[res] : 0.0;
	dd->rho += 0.5*violr[res]*violr[res];
	//fprintf(stderr,"right (%d): %7.3f up1: %7.3f low: %7.3f  viol: %7.3f\n",res,Rt_right[res],up1,low,violr[res]);fflush(stdout);

    }
    // Simple or DER-type ensemble: Rav is what we need to use
    else {
	Rav_6[res] = pow(Rav_6[res],-oneperexp);
	if (dd->dr_Mode == 2){
	    viol[res]=Rav_6[res] > up1 ? Rav_6[res]-up1 : Rav_6[res] < low ? low-Rav_6[res] : 0.0;
 	    //fprintf(stderr,"rav6 (%d): %7.3f up1: %7.3f low: %7.3f  viol: %7.3f\n",res,Rav_6[res],up1,low,viol[res]);fflush(stdout);

	    dd->rho += (1.0/dd->M)*viol[res]*viol[res];
	}
    }



     fa += 3*npairs;
     res++;
  }// while fa < nfa


    dd->rho/=(2.0*dd->nr);

    //if (MASTER(mcr))
     //fprintf(stderr,"HHDR> Rho = %7.5f, Rhomin = %7.5f\n",dd->rho,dd->rhomin);fflush(stderr);

     vtot=0.0;
    // No rhomin for the 1st calculation, so skip the main part
    if (dd->rhomin < 0.0){
        dd->rhomin=dd->rho;
    } else if (dd->rho > dd->rhomin){
	vtot=(dd->dr_fc*dd->M/2.0)*(dd->rho-dd->rhomin)*(dd->rho-dd->rhomin);
    
        // Here comes the force

	fa=0;
	res=0;

        // variables as in hqbm.src
	float pcor,pcor1,pcor2;
	float aux,aux1,aux2;
	float bux=2.0*(dd->rho-dd->rhomin)*dd->dr_fc/(float)dd->nr;
	float avdis,rnij;
	float fap;
	float onepersix=-1.0/6.0;

	while (fa < nfa){
	    type  = forceatoms[fa];
	    label = ip[type].hhdisres.label;
	    up1   = ip[type].hhdisres.up1;
	    up2   = ip[type].hhdisres.up2;
	    low   = ip[type].hhdisres.low;
	    k0    = smooth_fc*ip[type].hhdisres.kfac;

	    npairs=1;

            while(fa + 3*npairs < nfa &&
	          ip[forceatoms[fa + 3*npairs]].hhdisres.label == label)
		npairs++;

            // Variable name conversion :-)
	    rnij=pow(Rtl_6[res],-sixth);
            //rnij=RNij[res];

	    if (dd->dr_Mode == 3){
		if (dd->dr_exp == 6){
		    pcor1=pow(Rt_left[res]/rnij,7.0);
		    pcor2=pow(Rt_right[res]/rnij,7.0);
		}
		else if (dd->dr_exp == 3){
		    pcor1=pow(Rt_left[res]/rnij,4.0);
		    pcor2=pow(Rt_right[res]/rnij,4.0);
		}
		aux1=bux*pcor1*(violl[res])*pow(rnij,7.0);
		aux2=bux*pcor2*(violr[res])*pow(rnij,7.0);
		//fprintf(stderr,"rnij = %7.3f | pcor1 = %7.3f | pcor2 = %7.3f | aux1 = %7.3f | aux2 = %7.3f \n",rnij,pcor1,pcor2,aux1,aux2);fflush(stderr); 

	    }
	    else if (dd->dr_Mode == 2){
		if (dd->dr_exp == 6){
		    pcor=pow(Rav_6[res]/rnij,7.0);
		}
		else if (dd->dr_exp == 3){
		    pcor=pow(Rav_6[res]/rnij,4.0);
		}
		aux=bux*pcor*(viol[res])*pow(rnij,7.0);
	    }
	    else{
		aux=bux*(viol[res])*pow(rnij,7.0);
	    }

            /* Exert the force */
	    for(p=0; p<npairs; p++){
		pair = fa/3;
		ai   = forceatoms[fa+1];
		aj   = forceatoms[fa+2];

                // changing the direction:from ai-aj to aj-ai
		if (pbc)
		    ki=pbc_dx(pbc,x[aj],x[ai],dx);
                else
		    rvec_sub(x[aj],x[ai],dx);

		r=norm(dx);

		for(m=0; m<DIM; m++) {
                    // The exponent -8 is not yet clear for me
                    // MUMO-type force calc (pairwise averaging)
		    if (dd->dr_Mode == 3)
			fij            = aux1*dx[m]*pow(r,-8.0) + aux2*dx[m]*pow(r,-8.0);
	            else
			fij            = aux*dx[m]*pow(r,-8.0);

		    //fprintf(stderr,"RES: %d FORCE[%d]: %7.3f\n",res,m,fij);fflush(stderr); 

                    // I must try this
		    //fij*=-1;

		    f[ai][m]      += fij;
		    f[aj][m]      -= fij;
		    fshift[ki][m] += fij;
		    fshift[CENTRAL][m] -= fij;
		}

		fa += 3;
		     
            }// for pairs
	    res++;
	}// while fa < nfa  
	dd->sumviol = violtot;
    } //if rho > rhomin

    if (dd->rho < dd->rhomin)
	dd->rhomin=dd->rho;

    /* Return energy */
    return vtot;
 
}// ta_hhdisres
